import os
from typing import List, Optional

REDMAN_HISTORY = os.path.join(os.path.expanduser("~"), ".redman_history")


def create_history_file_default_if_not_exists() -> None:

    if not os.path.exists(REDMAN_HISTORY):
        with open(REDMAN_HISTORY, "a", encoding="UTF-8") as file:
            file.write("")


def add_history(redmine_name: str = None, status: str = "open", project_id: str = None, user_id: str = None) -> None:

    history = "\t".join([
        str(redmine_name) if redmine_name else "",
        str(status) if status else "",
        str(project_id) if project_id else "",
        str(user_id) if user_id else "",
    ]) + "\n"
    with open(REDMAN_HISTORY, "a", encoding="UTF-8") as file:
        file.write(history)


def load_history() -> Optional[List[str]]:

    if not os.path.exists(REDMAN_HISTORY):
        return None

    with open(REDMAN_HISTORY, "r", encoding="UTF-8") as file:
        return file.readlines()



from shutil import which
from typing import Optional

import click
from click.core import Context

from redman.config import (create_config_file_default_if_not_exists,
                           edit_config_file, load_config)
from redman.helpers.color import Color
from redman.helpers.fzf import fzf
from redman.helpers.redmine_api import (IssueStatus, UserStatus, list_issues,
                                        list_projects, list_users, show_issue,
                                        show_user)
from redman.helpers.table import Table
from redman.history import add_history, load_history


class Help:
    REDMINE_NAME = "redmine name. not project name. see your \"~/.redmanrc\". setup use \"redman config\". "
    ISSUE_STATUS = "choose from \"open\", \"close\", \"all\". default \"open\"."


REDMAN_COMMAND = "redman" if which("redman") else "python -m redman"


@click.group(invoke_without_command=True)
@click.pass_context
def redman(context: Context) -> None:
    if context.invoked_subcommand is None:
        history()


################################
# helper
################################
@redman.command()
def history() -> None:
    history_list = load_history()
    if not history_list:
        issues()
        return

    history_list.reverse()
    for history in history_list:
        if history.count("\t") == 3:
            redmine_name, status, project_id, user_id = history.split("\t")
            break

    issues(redmine_name if redmine_name else None,
           status if status else "open",
           project_id if project_id else None,
           user_id if user_id else None)


@redman.command()
def config() -> None:
    create_config_file_default_if_not_exists()

    edit_config_file()


################################
# projects
################################
@redman.command()
@click.option("--redmine_name", "-n", help=Help.REDMINE_NAME)
def projects(redmine_name: Optional[str]) -> None:

    url, api_key = load_config(redmine_name)
    if not url or not api_key:
        print("invalidate config")
        return

    body, err = list_projects(url, api_key)
    if err:
        print(err.reason)
        return

    projects = body.get("projects")
    if len(projects) <= 0:
        print("no project")
        return

    table = Table([
        "ID", "IDENTIFIER", "NAME", "DESCRIPTION",
    ])

    table.add_rows([[
        project.get("id"),
        project.get("identifier"),
        project.get("name"),
        project.get("description"),
    ] for project in projects])

    fzf(table.draw(),
        f"--bind=\"enter:abort+execute({REDMAN_COMMAND} issues --project_id={{1}} --redmine_name={redmine_name or ''} | more > /dev/tty)\"")


################################
# users
################################
@redman.command()
@click.option("--redmine_name", "-n", help=Help.REDMINE_NAME)
def users(redmine_name: Optional[str]) -> None:

    url, api_key = load_config(redmine_name)
    if not url or not api_key:
        print("invalidate config")
        return

    body, err = list_users(url, api_key)
    if err:
        print(err.reason)
        return

    users = body.get("users")
    if len(users) <= 0:
        print("no user")
        return

    table = Table([
        "ID", "LOGIN_ID", "NAME", "MAIL", "ADMIN", "LAST_LOGIN_ON",
    ])

    table.add_rows([[
        user.get("id"),
        user.get("login"),
        user.get("lastname") + " " + user.get("firstname"),
        user.get("mail"),
        "YES" if user.get("admin") else "",
        user.get("last_login_on"),
    ] for user in users])

    fzf(table.draw(),
        f"--preview=\"redman show user {{1}} --redmine_name={redmine_name or ''}\" \
          --bind=\"enter:abort+execute({REDMAN_COMMAND} issues --user_id={{1}} --redmine_name={redmine_name or ''} | more > /dev/tty)\"")


################################
# issues
################################
@redman.command()
@click.option("--redmine_name", "-n", help=Help.REDMINE_NAME)
@click.option("--status", "-s", type=click.Choice(["open", "close", "all"]), default="open", help=Help.ISSUE_STATUS)
@click.option("--project_id", "-p", type=str)
@click.option("--user_id", "-u", type=str)
def issues(redmine_name: Optional[str], status: str, project_id: Optional[str], user_id: Optional[str]) -> None:

    url, api_key = load_config(redmine_name)
    if not url or not api_key:
        print("invalidate config")
        return

    body, err = list_issues(url, api_key, IssueStatus.value_of(status), project_id, user_id)
    if err:
        print(err.reason)
        return

    issues = body.get("issues")
    if len(issues) <= 0:
        print("no issue")
        return

    table = Table([
        "ID", "TRACKER", "STATUS", "PRIORITY", "SUBJECT", "ASSIGNED", "DUE_DATE", "DESCRIPTION"
    ])

    table.add_rows([[
        issue.get("id"),
        issue.get("tracker").get("name"),
        issue.get("status").get("name"),
        issue.get("priority").get("name"),
        issue.get("subject"),
        issue.get("assigned_to", {}).get("name"),
        issue.get("due_date"),
        issue.get("description"),
    ] for issue in issues])

    stdin = table.draw()
    result = fzf(stdin,
                 f"--preview=\"{REDMAN_COMMAND} show issue {{1}} --redmine_name={redmine_name or ''}\"")
    if not result:
        return

    result = result.strip()

    add_history(redmine_name, status, project_id, user_id)

    for i, row in enumerate(stdin.splitlines()):
        row = row.strip()
        if row == result:
            print(f"""#{issues[i - 2].get("id")} {issues[i - 2].get("subject")}""")
            return


################################
# show
################################
@click.group()
def show() -> None:
    pass


@show.command()
@click.argument("id", type=str)
@click.option("--redmine_name", "-n", help=Help.REDMINE_NAME)
def user(id: str, redmine_name: Optional[str]) -> None:

    url, api_key = load_config(redmine_name)
    if not url or not api_key:
        print("invalidate config")
        return

    body, err = show_user(url, api_key, id)
    if err:
        print(err.reason)
        return

    user = body.get("user")
    preview = f"""{Color.背景緑} {Color.RESET} {user.get("lastname") + " " + user.get("firstname")}

{Color.灰}status     {Color.RESET}: {UserStatus.value_of(user.get("status")).name}
{Color.灰}login_id   {Color.RESET}: {user.get("login")}
{Color.灰}mail       {Color.RESET}: {user.get("mail")}
{Color.灰}api_key    {Color.RESET}: {user.get("api_key")}
{Color.灰}last_login {Color.RESET}: {user.get("last_login_on")}
{Color.灰}twofa      {Color.RESET}: {user.get("twofa_scheme")}

{Color.灰}groups     {Color.RESET}: {", ".join([group.get("name") for group in user.get("groups")])}
{Color.灰}memberships{Color.RESET}: {", ".join([group.get("project").get("name") for group in user.get("memberships")])}
    """

    print(preview)


@show.command()
@click.argument("id", type=str)
@click.option("--redmine_name", "-n", help=Help.REDMINE_NAME)
def issue(id: str, redmine_name: Optional[str]) -> None:

    url, api_key = load_config(redmine_name)
    if not url or not api_key:
        print("invalidate config")
        return

    body, err, = show_issue(url, api_key, id)
    if err:
        print(err.reason)
        return

    issue = body.get("issue")
    preview = f"""{issue.get("tracker").get("name")}

{Color.背景緑} {Color.RESET} {Color.緑}{Color.BOLD}#{issue.get("id")} {issue.get("subject")}{Color.RESET}
------------------------------------------------
 {issue.get("status").get("name")} | {issue.get("priority").get("name")} |
------------------------------------------------
{issue.get("description")}
    """
    print(preview)


redman.add_command(show)



import os
import subprocess
from typing import Optional, Tuple

import yaml

REDMANRC = os.path.join(os.path.expanduser("~"), ".redmanrc")


def create_config_file_default_if_not_exists() -> None:

    if not os.path.exists(REDMANRC):
        config_file_default = {
            "default": "my-redmine",
            "my-redmine": {
                "REDMINE_URL": "<your Redmine url>",
                "REDMINE_API_ACCESS_KEY": "<your Redmine Access Key>"
            }
        }
        with open(REDMANRC, "a", encoding="UTF-8") as config_file:
            yaml.dump(config_file_default, config_file)


def edit_config_file() -> None:
    subprocess.run(
        f"eval ${{EDITOR:-vi}} {REDMANRC}",
        shell=True, check=False)


def load_config(redmine_name: Optional[str]) -> Tuple[Optional[str], Optional[str]]:

    if not os.path.exists(REDMANRC):
        return None, None

    with open(REDMANRC, "r", encoding="UTF-8") as config_file:
        config = yaml.safe_load(config_file)

    if not redmine_name:
        redmine_name = config.get("default")
        if not redmine_name:
            return None, None

    obj = config.get(redmine_name)

    if not obj:
        return None, None

    return obj.get("REDMINE_URL"), obj.get("REDMINE_API_ACCESS_KEY")



import subprocess

FZF_OPTIONS = "--header-lines=2 \
    --ansi \
    --no-multi \
    --no-mouse \
    --cycle \
    --reverse \
    --bind=alt-p:toggle-preview \
    --bind=alt-w:toggle-preview-wrap \
    --bind=ctrl-w:abort \
    --bind=shift-down:preview-down \
    --bind=shift-up:preview-up \
    --bind=shift-down:preview-down \
    --bind=shift-left:preview-page-up \
    --bind=shift-right:preview-page-down \
    --preview-window=down:50%:rounded:wrap"


def fzf(stdin: str, options: str = "") -> str:
    return subprocess.run(
        f"echo \"{stdin}\" | fzf {FZF_OPTIONS} {options}",
        shell=True, check=False, stdout=subprocess.PIPE) \
        .stdout.decode("utf-8").strip()



from __future__ import annotations

from typing import Any, Iterable
from texttable import Texttable


class Table:

    def __init__(self, header: list[str]) -> None:
        self.table = Texttable(max_width=0)
        self.table.set_deco(Texttable.HEADER | Texttable.VLINES)
        self.table.header(header)

    def add_rows(self, rows: Iterable[list[Any]]) -> None:
        # self.table.add_rows(rows)
        self.table.add_rows([[str(cell).replace("\n", " ") if cell else "" for cell in row] for row in rows], header=False)

    def draw(self) -> str:
        return self.table.draw()



from __future__ import annotations
# https://stackoverflow.com/questions/33533148/how-do-i-type-hint-a-method-with-the-type-of-the-enclosing-class

import json
from enum import Enum
from typing import Any, Optional, Tuple
from urllib import error, parse, request


################################
# projects
################################
def list_projects(base_url: str, api_key: str) -> Tuple[Any, Optional[error.URLError]]:
    """GET /projects.[format]"""
    try:
        params = {
            "key": api_key
        }
        req = request.Request(f"{base_url}/projects.json?{parse.urlencode(params)}")

        with request.urlopen(req) as res:
            return json.load(res), None
    except error.URLError as err:
        return None, err


################################
# users
################################
class UserStatus(Enum):
    ANONYMOUS = 0
    ACTIVE = 1
    REGISTERED = 2
    LOCKED = 3
    UNKNOWN = -1

    @classmethod
    def value_of(cls, value: Optional[int]) -> UserStatus:
        if not value:
            return UserStatus.UNKNOWN

        for e in UserStatus:
            if value == e.value:
                return e

        return UserStatus.UNKNOWN


def list_users(base_url: str, api_key: str) -> Tuple[Any, Optional[error.URLError]]:
    """GET /users.json"""
    try:
        params = {
            "key": api_key
        }
        req = request.Request(f"{base_url}/users.json?{parse.urlencode(params)}")

        with request.urlopen(req) as res:
            return json.load(res), None
    except error.URLError as err:
        return None, err


def show_user(base_url: str, api_key: str, id: str) -> Tuple[Any, Optional[error.URLError]]:
    """GET /users/[id].[format]"""
    try:
        params = {
            "key": api_key,
            "include": "memberships,groups",
        }

        req = request.Request(f"{base_url}/users/{id}.json?{parse.urlencode(params)}")

        with request.urlopen(req) as res:
            return json.load(res), None
    except error.URLError as err:
        return None, err


################################
# issues
################################
class IssueStatus(Enum):
    OPEN = "o", "o"
    CLOSED = "c", "c"
    ALL = "*", "a"

    @classmethod
    def value_of(cls, value: Optional[str]) -> IssueStatus:
        if not value:
            return IssueStatus.OPEN

        for e in IssueStatus:
            if value.startswith(e.value[1]):
                return e

        return IssueStatus.OPEN


def list_issues(base_url: str, api_key: str,
                status: IssueStatus = IssueStatus.OPEN, project_id: str = None, user_id: str = None
                ) -> Tuple[Any, Optional[error.URLError]]:
    """GET /issues.[format]"""
    try:
        params = {
            "key": api_key,
            "sort": "due_date,id",
            "f[]": "status_id",
            "op[status_id]": status.value[0],
        }

        if project_id:
            params["f[]"] = "project_id"
            params["op[project_id]"] = "="
            params["v[project_id][]"] = project_id

        if user_id:
            params["f[]"] = "assigned_to_id"
            params["op[assigned_to_id]"] = "="
            params["v[assigned_to_id][]"] = user_id

        url = f"{base_url}/issues.json?{parse.urlencode(params)}"

        req = request.Request(url)

        with request.urlopen(req) as res:
            return json.load(res), None
    except error.URLError as err:
        return None, err


def show_issue(base_url: str, api_key: str, id: str) -> Tuple[Any, Optional[error.URLError]]:
    """GET /issues/[id].[format]"""
    try:
        params = {
            "key": api_key,
        }

        req = request.Request(f"{base_url}/issues/{id}.json?{parse.urlencode(params)}")

        with request.urlopen(req) as res:
            return json.load(res), None
    except error.URLError as err:
        return None, err
